/*
 * JZ4780 HDMI peripheral support.
 *
 * Copyright (C) 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * Some structures have been adopted from the Linux DRM bridge driver for
 * Synopsys DW-HDMI with the following attribution:
 *
 * Copyright (C) 2011 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/devices/lcd-jz4740-panel.h>
#include <l4/sys/types.h>
#include <stdint.h>



// PHY capability definition type.

struct Phy_capabilities
{
  uint8_t type;
  const char *name;
  int gen;
  int svsret;
  int configure;
};



// PHY configuration types.

enum Phy_resolutions
{
  Phy_resolution_8bpc = 0,
  Phy_resolution_10bpc = 1,
  Phy_resolution_12bpc = 2,
  Phy_resolution_count = 3,
};

struct Phy_mpll_config
{
  unsigned long pixelclock; // frequency (Hz)
  struct {
    uint16_t cpce;
    uint16_t gmp;
  } res[Phy_resolution_count];
};

struct Phy_curr_ctrl
{
  unsigned long pixelclock; // frequency (Hz)
  uint16_t curr[Phy_resolution_count];
};

struct Phy_config
{
  unsigned long pixelclock; // frequency (Hz)
  uint16_t symbol;          // clock symbol and transmitter control
  uint16_t term;            // transmission termination value
  uint16_t vlevel;          // voltage level control
};



#ifdef __cplusplus

#include <l4/devices/hw_mmio_register_block.h>

// HDMI device control.

class Hdmi_jz4780_chip
{
private:
  l4_addr_t _start, _end;
  l4_cap_idx_t _irq;
  Hw::Register_block<8> _regs;

  // Identification.

  uint16_t _version;
  uint8_t _phy_type;
  const struct Phy_capabilities *_phy_def;

  // Input/output properties.

  struct Jz4740_lcd_panel *_panel;
  unsigned long _pixelclock;

  // Transfer properties.

  bool _segment_read;
  uint8_t _device_register, _phy_device_register;

  // Convenience methods for register access.

  void reg_update(uint32_t reg, uint32_t bits, bool enable);
  void reg_update_field(uint32_t reg, uint32_t mask, uint32_t bits);
  void reg_fill_field(uint32_t reg, uint32_t mask);

protected:
  unsigned long get_pixelclock();

  void get_identification();

  void irq_init();

  void i2c_init(uint32_t reset, uint32_t divider, uint32_t config0,
                uint32_t config1, uint32_t status, uint32_t mask);

  long i2c_wait(uint32_t status);

  void phy_irq_init();

public:
  Hdmi_jz4780_chip(l4_addr_t start, l4_addr_t end, l4_cap_idx_t irq,
                   struct Jz4740_lcd_panel *panel);

  // Chipset querying.

  void get_version(uint8_t *major, uint16_t *minor);
  void get_phy_capabilities(const struct Phy_capabilities **phy_def);

  // I2C operations.

  int i2c_read(uint8_t *buf, unsigned int length);

  void i2c_set_address(uint8_t address);
  void i2c_set_segment(uint8_t segment);
  void i2c_set_register(uint8_t device_register);

  int i2c_phy_write(uint8_t address, uint16_t value);
  int i2c_phy_write(uint16_t *buf, unsigned int length);

  void i2c_phy_set_address(uint8_t address);
  void i2c_phy_set_register(uint8_t device_register);

  // PHY configuration operations.

  void phy_enable_powerdown(bool enable);
  void phy_enable_tmds(bool enable);
  void phy_enable_svsret(bool enable);
  void phy_enable_gen2_powerdown(bool enable);
  void phy_enable_gen2_tx_power(bool enable);
  void phy_enable_interface(bool enable);

  // PHY operations.

  long phy_configure();
  long phy_configure_specific();
  long phy_init();
  void phy_power_off();
  void phy_power_on();
  void phy_reset();

  // Hotplug and signalling support.

  bool connected();
  long wait_for_connection();
  long wait_for_tx_phy_lock(int level);
  long wait_for_phy_irq(uint32_t int_status_flags, uint32_t status_flags,
                        uint32_t polarity_flags);

  // Video output initialisation.

  void enable_overflow_irq(bool enable);
  void frame_init();
  void data_path_init();
  void packet_init();
  void csc_init();
  void sample_init();
  void hdcp_init();

  // Enable the video output.

  long enable(unsigned long pixelclock);
};

#endif /* __cplusplus */



/* C language interface. */

EXTERN_C_BEGIN

void *jz4780_hdmi_init(l4_addr_t start, l4_addr_t end, l4_cap_idx_t irq,
                       struct Jz4740_lcd_panel *panel);

void jz4780_hdmi_get_version(void *hdmi, uint8_t *major, uint16_t *minor);

void jz4780_hdmi_get_phy_capabilities(void *hdmi, const struct Phy_capabilities **phy_def);

int jz4780_hdmi_i2c_read(void *hdmi, uint8_t *buf, unsigned int length);

void jz4780_hdmi_i2c_set_address(void *hdmi, uint8_t address);

void jz4780_hdmi_i2c_set_segment(void *hdmi, uint8_t segment);

void jz4780_hdmi_i2c_set_register(void *hdmi, uint8_t device_register);

int jz4780_hdmi_phy_i2c_write(void *hdmi, uint16_t *buf, unsigned int length);

void jz4780_hdmi_i2c_phy_set_address(void *hdmi, uint8_t address);

void jz4780_hdmi_i2c_phy_set_register(void *hdmi, uint8_t device_register);

int jz4780_hdmi_connected(void *hdmi);

long jz4780_hdmi_wait_for_connection(void *hdmi);

long jz4780_hdmi_enable(void *hdmi, unsigned long pixelclock);

EXTERN_C_END
